@@ -1,11 +1,8 @@
 /*
  * BIOS Decode
  *
- *      (C) 2000-2002 Alan Cox <alan@redhat.com>
- *      (C) 2002-2003 Jean Delvare <khali@linux-fr>
- *
- * Licensed under the GNU Public license. If you want to use it in with
- * another license just ask.
+ *   (C) 2000-2002 Alan Cox <alan@redhat.com>
+ *   (C) 2002-2003 Jean Delvare <khali@linux-fr>
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -46,6 +43,9 @@
  *  - Microsoft "PCI IRQ Routing Table Specification"
  *    Version 1.0
  *    http://www.microsoft.com/hwdev/archive/BUSBIOS/pciirq.asp
+ *  - Compaq "Technical Reference Guide for Compaq Deskpro 4000 and 6000"
+ *    First Edition
+ *    http://h18000.www1.hp.com/support/techpubs/technical_reference_guides/113a1097.html
  */
 
 #include <sys/types.h>
@@ -55,14 +55,10 @@
 #include <fcntl.h>
 #include <string.h>
 #include <stdlib.h>
-#include <errno.h>
-#include <ctype.h>
 
 #include "version.h"
-
-typedef unsigned char u8;
-typedef unsigned short u16;
-typedef unsigned int u32;
+#include "types.h"
+#include "util.h"
 
 #ifdef BIGENDIAN
 typedef struct {
@@ -88,50 +84,6 @@ struct bios_entry {
 	int (*decode)(const u8*, size_t);
 };
 
-/*
- * Tools
- */
-
-static int myread(int fd, u8 *buf, size_t count, const char *prefix)
-{
-	ssize_t r=1;
-	size_t r2=0;
-	
-	while(r2!=count && r!=0)
-	{
-		r=read(fd, buf+r2, count-r2);
-		if(r==-1)
-		{
-			if(errno!=EINTR)
-			{
-				close(fd);
-				perror(prefix);
-				return -1;
-			}
-		}
-		else
-			r2+=r;
-	}
-	
-	if(r2!=count)
-	{
-		close(fd);
-		fprintf(stderr, "%s: Unexpected end of file\n", prefix);
-		return -1;
-	}
-	
-	return 0;
-}
-
-static __inline__ int checksum(const u8 *buf, size_t len)
-{
-	u8 sum=0;
-	size_t a;
-	
-	for(a=0; a<len; a++)
-		sum+=buf[a];
-	return (sum==0);
-}
 
 /*
  * SMBIOS
@@ -437,6 +389,44 @@ static int pir_decode(const u8 *p, size_t len)
 	return 1;
 }
 
+/*
+ * Compaq-specific entries
+ */
+
+static size_t compaq_length(const u8 *p)
+{
+	return (p[4]*10+5);
+}
+
+static int compaq_decode(const u8 *p, __attribute__ ((unused)) size_t len)
+{
+	int i;
+
+	printf("Compaq-specific entries present.\n");
+
+	/* integrity checking (lack of checksum) */
+	for(i=0; i<p[4]; i++)
+	{
+		if(p[5+i*10]!='$' || !(p[6+i*10]>='A' && p[6+i*10]<='Z')
+			|| !(p[7+i*10]>='A' && p[7+i*10]<='Z')
+			|| !(p[8+i*10]>='A' && p[8+i*10]<='Z'))
+		{
+			printf("\t Abnormal Entry! Please report. [%02X %02X %02X %02X]\n",
+				p[5+i*10], p[6+i*10], p[7+i*10], p[8+i*10]);
+			return 0;
+		}
+	}
+	
+	for(i=0; i<p[4]; i++)
+	{
+		printf("\tEntry %u: %c%c%c%c at 0x%08X (%u bytes)\n",
+			i+1, p[5+i*10], p[6+i*10], p[7+i*10], p[8+i*10],
+			DWORD(p+9+i*10), WORD(p+13+i*10));
+	}
+
+	return 1;
+}
+
 /*
  * Main
  */
@@ -450,6 +440,7 @@ static struct bios_entry bios_entries[]={
 	{ "$SNY", 0xE0000, 0xFFFFF, sony_length, sony_decode },
 	{ "_32_", 0xE0000, 0xFFFFF, bios32_length, bios32_decode },
 	{ "$PIR", 0xF0000, 0xFFFFF, pir_length, pir_decode },
+	{ "32OS", 0xE0000, 0xFFFFF, compaq_length, compaq_decode },
 	{ NULL, 0, 0, NULL, NULL }
 };
 
