@@ -1,12 +1,8 @@
 /*
  * DMI Decode
  *
- *      (C) 2000-2002 Alan Cox <alan@redhat.com>
- *      (C) 2002-2003 Jean Delvare <khali@linux-fr>
- *
- *
- * Licensed under the GNU Public license. If you want to use it in with
- * another license just ask.
+ *   (C) 2000-2002 Alan Cox <alan@redhat.com>
+ *   (C) 2002-2003 Jean Delvare <khali@linux-fr>
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -50,23 +46,22 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#ifdef USE_MMAP
+#include <sys/mman.h>
+#endif /* USE MMAP */
 #include <stdio.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <string.h>
 #include <stdlib.h>
-#include <errno.h>
 
 #include "version.h"
+#include "types.h"
+#include "util.h"
 
 static const char *out_of_spec = "<OUT OF SPEC>";
 static const char *bad_index = "<BAD INDEX>";
 
-typedef unsigned char u8;
-typedef unsigned short u16;
-typedef signed short i16;
-typedef unsigned int u32;
-
 /*
  * The specification isn't very clear on endianness problems, so we better
  * have macros for these. It also helps us solve problems on systems that
@@ -145,40 +140,6 @@ struct dmi_header
 #define HANDLE(x) x->handle
 #endif
 
-/*
- * Tools
- */
-
-static int myread(int fd, u8 *buf, size_t count, const char *prefix)
-{
-	ssize_t r=1;
-	size_t r2=0;
-	
-	while(r2!=count && r!=0)
-	{
-		r=read(fd, buf+r2, count-r2);
-		if(r==-1)
-		{
-			if(errno!=EINTR)
-			{
-				close(fd);
-				perror(prefix);
-				return -1;
-			}
-		}
-		else
-			r2+=r;
-	}
-	
-	if(r2!=count)
-	{
-		close(fd);
-		fprintf(stderr, "%s: Unexpected end of file\n", prefix);
-		return -1;
-	}
-	
-	return 0;
-}
 
 /*
  * Type-independant Stuff
@@ -977,7 +938,7 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		/*
 		 * 80386 have a different signature.
 		 */
-		printf("%sSignature: Type %u, Family %u, Major Stepping %u, Minor Stepping %u\n",
+		printf("%sSignature: Type %X, Family %X, Major Stepping %X, Minor Stepping %X\n",
 			prefix, dx>>12, (dx>>8)&0xF, (dx>>4)&0xF, dx&0xF);
 		return;
 	}
@@ -1013,7 +974,7 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		return;
 	
 	eax=DWORD(p);
-	printf("%sSignature: Type %u, Family %u, Model %u, Stepping %u\n",
+	printf("%sSignature: Type %X, Family %X, Model %X, Stepping %X\n",
 		prefix, (eax>>12)&0x3, ((eax>>16)&0xFF0)+((eax>>8)&0x00F),
 		((eax>>12)&0xF0)+((eax>>4)&0x0F), eax&0xF);
 	if(cpuid)
@@ -3679,10 +3640,7 @@ static void dmi_decode(u8 *data, u16 ver)
 		case 38: /* 3.3.39 IPMI Device Information */
 			/*
 			 * We use the word "Version" instead of "Revision", conforming to
-			 * the IPMI 1.5 specification. This specification isn't very clear
-			 * regarding the I2C slave address. I couldn't understand wether
-			 * or not we are supposed to shift it by one bit to the right, so
-			 * I leave it untouched. Beware it might be wrong.
+			 * the IPMI 1.5 specification.
 			 */
 			printf("\tIPMI Device Information\n");
 			if(h->length<0x10) break;
@@ -3691,7 +3649,7 @@ static void dmi_decode(u8 *data, u16 ver)
 			printf("\t\tSpecification Version: %u.%u\n",
 				data[0x05]>>4, data[0x05]&0x0F);
 			printf("\t\tI2C Slave Address: 0x%02x\n",
-				data[0x06]);
+				data[0x06]>>1);
 			if(data[0x07]!=0xFF)
 				printf("\t\tNV Storage Device Address: %u\n",
 					data[0x07]);
@@ -3789,17 +3747,43 @@ static void dmi_decode(u8 *data, u16 ver)
 	}
 }
 		
+#ifdef USE_MMAP
+static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, __attribute__ ((unused)) const char *pname, __attribute__ ((unused)) const char *devmem)
+#else /* USE_MMAP */
 static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, const char *pname, const char *devmem)
+#endif /* USE_MMAP */
 {
 	u8 *buf;
 	u8 *data;
 	int i=0;
+#ifdef USE_MMAP
+	div_t mmbase;
+	size_t psize;
+	void *mmp;
+#endif /* USE_MMAP */
 	
 	printf("%u structures occupying %u bytes.\n",
 		num, len);
 	printf("Table at 0x%08X.\n",
 		base);
 	
+#ifdef USE_MMAP
+	psize=getpagesize();
+	mmbase=div(base, psize);
+	/*
+	 * We need PROT_WRITE for ASCII filtering in dmi_string. Do NOT change MAP_PRIVATE
+	 * to MAP_SHARED unless you also remove PROT_WRITE and disable ASCII filtering, because
+	 * we DO NOT WANT to write to /dev/mem. Thanks a lot to Gabriel Forte for helping
+	 * me hunting down an annoying bug related to this.
+	 */
+	mmp=mmap(0, mmbase.rem+len, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, mmbase.quot*psize);
+    if(mmp==MAP_FAILED)
+    {
+       perror("mmap");
+       return;
+    }
+	buf=(u8 *)mmp+mmbase.rem;
+#else /* USE_MMAP */
 	if((buf=malloc(len))==NULL)
 	{
 		perror(pname);
@@ -3813,11 +3797,12 @@ static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, const char *p
 	if(myread(fd, buf, len, devmem)==-1)
 	{
 		free(buf);
-		printf("Table is unreachable, sorry. This problem is known on IBM T23 and T30 laptops,\n"
-			"as well as on the Fujitsu-Siemens S-4582 laptop. If your system differ, please\n"
-			"report.\n");
+		printf("Table is unreachable, sorry. Try compiling dmidecode with -DUSE_MMAP.\n"
+			"This problem is known on the IBM T23, T30 and X30 laptops, the Fujitsu-Siemens\n"
+			"S-4582 laptop as well as IA-64 systems. If your system differ, please report\n");
 		exit(1);
 	}
+#endif /* USE_MMAP */
 	
 	data=buf;
 	while(i<num && data+sizeof(struct dmi_header)<=buf+len)
@@ -3835,6 +3820,8 @@ static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, const char *p
 		next+=2;
 		if(next-buf<=len)
 			dmi_decode(data, ver);
+		else
+			printf("\t<TRUNCATED>\n");
 		
 		data=next;
 		i++;
@@ -3844,28 +3831,52 @@ static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, const char *p
 		printf("Wrong DMI structures count: %d announced, only %d decoded.\n",
 			num, i);
 	if(data-buf!=len)
-		printf("Wrong DMI structures length: %d bytes announced, %d bytes decoded.\n",
-			len, data-buf);
+		printf("Wrong DMI structures length: %d bytes announced, structures occupy %d bytes.\n",
+			len, (unsigned int)(data-buf));
 	
+#ifdef USE_MMAP
+	if(munmap(mmp, mmbase.rem+len)==-1)
+		perror("munmap");
+#else /* USE_MMAP */
 	free(buf);
+#endif /* USE_MMAP */
 }
 
 
-static int checksum(u8 *buf, u8 len)
+static int smbios_decode(u8 *buf, int fd, const char *pname, const char *devmem)
 {
-	u8 sum=0, a;
+	if(checksum(buf, buf[0x05])
+	 && memcmp(buf+0x10, "_DMI_", 5)==0
+	 && checksum(buf+0x10, 0x0F))
+	{
+		printf("SMBIOS %u.%u present.\n",
+			buf[0x06], buf[0x07]);
+		dmi_table(fd, DWORD(buf+0x18), WORD(buf+0x16), WORD(buf+0x1C),
+			(buf[0x06]<<8)+buf[0x07], pname, devmem);
+		return 1;
+	}
 	
-	for(a=0; a<len; a++)
-		sum+=buf[a];
-	return (sum==0);
+	return 0;
 }
 
 int main(int argc, const char *argv[])
 {
-	u8 buf[0x20];
 	int fd;
 	off_t fp=0xF0000;
 	const char *devmem="/dev/mem";
+#ifdef __IA64__
+	FILE *efi_systab;
+	char linebuf[64];
+#ifdef USE_MMAP
+	div_t mmbase;
+	size_t psize;
+	void *mmp;
+#else /* USE_MMAP */
+	u8 buf[0x20];
+#endif /* USE_MMAP */
+#else /* __IA64__ */
+	u8 buf[0x20];
+#endif /* __IA64__ */
 	
 	if(sizeof(u8)!=1 || sizeof(u16)!=2 || sizeof(u32)!=4 || '\0'!=0)
 	{
@@ -3875,34 +3886,78 @@ int main(int argc, const char *argv[])
 	
 	if(argc>=2)
 		devmem=argv[1];
-	if((fd=open(devmem, O_RDONLY))==-1 || lseek(fd, fp, SEEK_SET)==-1)
+	if((fd=open(devmem, O_RDONLY))==-1)
 	{
 		perror(devmem);
 		exit(1);
 	}
 	
 	printf("# dmidecode %s\n", VERSION);
-	while(fp<0xFFFFF)
+	
+#ifdef __IA64__
+	if((efi_systab=fopen("/proc/efi/systab", "r"))==NULL)
+	{
+		perror("/proc/efi/systab");
+		exit(1);
+	}
+	while((fgets(linebuf, sizeof(linebuf)-1, efi_systab))!=NULL)
+	{
+		char* addr=memchr(linebuf, '=', strlen(linebuf));
+		*(addr++)='\0';
+		if(strcmp(linebuf, "SMBIOS")==0)
+		{
+			fp=strtol(addr, NULL, 0);
+			printf("# SMBIOS entry point at 0x%08lx\n", fp);
+		}
+	}
+	if(fclose(efi_systab)!=0)
+		perror("/proc/efi/systab");
+
+#ifdef USE_MMAP
+	psize=getpagesize();
+	mmbase=div(fp, psize);
+	mmp=mmap(0, mmbase.rem+0x20, PROT_READ, MAP_PRIVATE, fd, mmbase.quot*psize);
+    if(mmp==MAP_FAILED)
+    {
+       perror(devmem);
+       exit(1);
+    }
+
+	smbios_decode(((u8 *)mmp)+mmbase.rem, fd, argv[0], devmem);
+
+	if(munmap(mmp, mmbase.rem+0x20)==-1)
+		perror("munmap");
+#else /* USE_MMAP */
+	if(lseek(fd, fp, SEEK_SET)==-1)
+	{
+		perror(devmem);
+		exit(1);
+	}
+	if(myread(fd, buf, 0x20, devmem)==-1)
+		exit(1);
+
+	smbios_decode(buf, fd, argv[0], devmem);
+#endif /* USE_MMAP */
+#else /* __IA64__ */
+	if(lseek(fd, fp, SEEK_SET)==-1)
+	{
+		perror(devmem);
+		exit(1);
+	}
+	while(fp<=0xFFFF0)
 	{
 		if(myread(fd, buf, 0x10, devmem)==-1)
 			exit(1);
 		fp+=16;
 		
-		if(memcmp(buf, "_SM_", 4)==0 && fp<0xFFFFF)
+		if(memcmp(buf, "_SM_", 4)==0 && fp<=0xFFFF0)
 		{
 			if(myread(fd, buf+0x10, 0x10, devmem)==-1)
 				exit(1);
 			fp+=16;
 			
-			if(checksum(buf, buf[0x05])
-			 && memcmp(buf+0x10, "_DMI_", 5)==0
-			 && checksum(buf+0x10, 0x0F))
+			if(smbios_decode(buf, fd, argv[0], devmem))
 			{
-				printf("SMBIOS %u.%u present.\n",
-					buf[0x06], buf[0x07]);
-				dmi_table(fd, DWORD(buf+0x18), WORD(buf+0x16), WORD(buf+0x1C),
-					(buf[0x06]<<8)+buf[0x07], argv[0], devmem);
-				
 				/* dmi_table moved us far away */
 				lseek(fd, fp, SEEK_SET);
 			}
@@ -3919,6 +3974,7 @@ int main(int argc, const char *argv[])
 			lseek(fd, fp, SEEK_SET);
 		}
 	}
+#endif /* __IA64__ */
 	
 	if(close(fd)==-1)
 	{
